/***********************************************************

Copyright (c) 1987  X Consortium

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of the X Consortium shall not be
used in advertising or otherwise to promote the sale, use or other dealings
in this Software without prior written authorization from the X Consortium.


Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of Digital not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/
/* $XConsortium: mfbpixmap.c,v 5.13 94/04/17 20:28:28 dpw Exp $ */

/************************************************************
Copyright (c) 1994  Sun Microsystems, Inc.
Copyright (c) 1993  X Consortium

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documenta-
tion files (the "Software"), to deal in the Software with-
out restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to
whom the Software is furnished to do so, subject to the 
following conditions:

The above copyright notice and this permission notice shall
be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE, AND NONINFRINGEMENT.  IN NO EVENT SHALL SUN OR THE
X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES, OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHER-
WISE, ARISING FROM, OUT OF, OR IN CONNECTION WITH THE 
SOFTWARE OR THE USE OR DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the names of Sun
Microsystems, Inc. and the X Consortium shall not be used in
advertising or otherwise to promote the sale, use, or other
dealings in this Software without prior written authoriza-
tion from Sun Microsystems, Inc. and/or the X Consortium, as
appropriate.
*************************************************************/

/*
 * modified for X11R6 Xvan server
 */

/* pixmap management
   written by drewry, september 1986

   on a monchrome device, a pixmap is a bitmap.
*/

extern "C" {
#include "Xmd.h"
#include "scrnintstr.h"
#include "pixmapstr.h"
#include "mi.h"
#include "servermd.h"
}

#include "mfb.h"
#include "maskbits.h"

PixmapPtr
mfbCreatePixmap (ScreenPtr pScreen, int width, int height, int depth)
{
  PixmapPtr pPixmap;
  FBList *fblistp;
  int datasize;
  int paddedWidth;
  
  if (depth != 1)
    return NullPixmap;
  paddedWidth = BitmapBytePad(width);
  datasize = height * paddedWidth;
  pPixmap = AllocatePixmap(pScreen, sizeof(FBList) + datasize);
  if (!pPixmap)
    return NullPixmap;
  pPixmap->drawable.type = DRAWABLE_PIXMAP; 
  pPixmap->drawable.class = 0;
  pPixmap->drawable.pScreen = pScreen;
  pPixmap->drawable.depth = depth;
  pPixmap->drawable.bitsPerPixel = depth;
  pPixmap->drawable.id = 0;
  pPixmap->drawable.serialNumber = NEXT_SERIAL_NUMBER;
  pPixmap->drawable.x = 0;
  pPixmap->drawable.y = 0;
  pPixmap->drawable.width = width;
  pPixmap->drawable.height = height;
  pPixmap->devKind = paddedWidth;
  pPixmap->refcnt = 1;
  
#ifdef PIXPRIV
  fblistp = new((void *) ((char *)pPixmap + pScreen->totalPixmapSize)) FBList;
#else
  fblistp = new((void *) (pPixmap + 1)) FBList;
#endif

  if (datasize)
    fblistp->add(width, height, depth, (unsigned int *)(fblistp + 1));
  pPixmap->devPrivate.ptr = (pointer) fblistp;
  return pPixmap;
}

Bool
mfbDestroyPixmap(PixmapPtr pPixmap)
{
  FBList *fblistp;

  if(--pPixmap->refcnt)
    return TRUE;

#ifdef PIXPRIV
  fblistp = (FBList *) ((char *) pPixmap + pScreen->totalPixmapSize);
#else
  fblistp = (FBList *) (pPixmap + 1);
#endif
  fblistp->~FBList();
  xfree(pPixmap);
  return TRUE;
}


PixmapPtr
mfbCopyPixmap(register PixmapPtr pSrc)
{
    register PixmapPtr	pDst;
    int		size;
    ScreenPtr pScreen;

    size = pSrc->drawable.height * pSrc->devKind;
    pScreen = pSrc->drawable.pScreen;
    pDst = (*pScreen->CreatePixmap) (pScreen, pSrc->drawable.width, 
				pSrc->drawable.height, pSrc->drawable.depth);
    if (!pDst)
	return NullPixmap;
    FBList *fblistsrc = (FBList *) pSrc->devPrivate.ptr;
    FBList *fblistdst = (FBList *) pDst->devPrivate.ptr;
    memmove((char *)fblistdst->getAddr(), (char *)fblistsrc->getAddr(), size);
    return pDst;
}


/* replicates a pattern to be a full 32 bits wide.
   relies on the fact that each scnaline is longword padded.
   doesn't do anything if pixmap is not a factor of 32 wide.
   changes width field of pixmap if successful, so that the fast
	XRotatePixmap code gets used if we rotate the pixmap later.

   calculate number of times to repeat
   for each scanline of pattern
      zero out area to be filled with replicate
      left shift and or in original as many times as needed
*/
void
mfbPadPixmap(PixmapPtr pPixmap)
{
    register int width = pPixmap->drawable.width;
    register int h;
    register PixelType mask;
    FBP p;
    register PixelType bits;	/* real pattern bits */
    register int i;
    int rep;			/* repeat count for pattern */

    if (width >= PPW)
	return;

    rep = PPW/width;
    if (rep*width != PPW)
	return;

    mask = endtab[width];

    p = (FBList *)(pPixmap->devPrivate.ptr);
    for (h=0; h < pPixmap->drawable.height; h++)
    {
	*p &= mask;
	bits = *p;
	for(i=1; i<rep; i++)
	{
	    bits = SCRRIGHT(bits, width);
	    *p |= bits;
	}
	p++;
    }
    pPixmap->drawable.width = PPW;
}

/* Rotates pixmap pPix by w pixels to the right on the screen. Assumes that
 * words are PPW bits wide, and that the least significant bit appears on the
 * left.
 */
void
mfbXRotatePixmap(PixmapPtr pPix, register int rw)
{
    FBP pw, pwFinal;
    register PixelType	t;

    if (pPix == NullPixmap)
        return;

    pw = (FBList *) pPix->devPrivate.ptr;
    rw %= (int)pPix->drawable.width;
    if (rw < 0)
	rw += (int)pPix->drawable.width;
    if(pPix->drawable.width == PPW)
    {
        pwFinal = pw + pPix->drawable.height;
	while(pw < pwFinal)
	{
	    t = *pw;
	    *pw++ = SCRRIGHT(t, rw) | 
		    (SCRLEFT(t, (PPW-rw)) & endtab[rw]);
	}
    }
    else
    {
	/* We no longer do this.  Validate doesn't try to rotate odd-size
	 * tiles or stipples.  mfbUnnatural<tile/stipple>FS works directly off
	 * the unrotate tile/stipple in the GC
	 */
        ErrorF("X internal error: trying to rotate odd-sized pixmap.\n");
    }

}

/* Rotates pixmap pPix by h lines.  Assumes that h is always less than
   pPix->height
   works on any width.
 */
void
mfbYRotatePixmap(register PixmapPtr pPix, int rh)
{
    int nlDown;	        /* longs to move down to row 0; also offset of
			   row rh */
    int nlUp;		/* longs to move up to line rh; also
			   offset of first line moved down to 0 */
    FBP pbase, ps, pd;
    unsigned long *ptmp, *pt;
    int	height, i;

    if (pPix == NullPixmap)
	return;
    height = (int) pPix->drawable.height;
    rh %= height;
    if (rh < 0)
	rh += height;

    pbase = (FBList *) pPix->devPrivate.ptr;

    nlDown = rh * (pPix->devKind >> 2);
    nlUp = ((pPix->devKind >> 2) * height) - nlDown;
    if(!(ptmp = (unsigned long *)ALLOCATE_LOCAL(nlUp << 2)))
	return;

    ps = pbase;
    pt = ptmp;
    for (i = 0; i < nlUp; ++ i)
      *(pt ++) = *(ps ++);

    ps = pbase + nlUp;
    pd = pbase;
    for (i = 0; i < nlDown; ++ i)
      *(pd ++) = *(ps ++);

    pt = ptmp;
    pd = pbase + nlDown;
    for (i = 0; i < nlUp; ++ i)
      *(pd ++) = *(pt ++);

    DEALLOCATE_LOCAL(ptmp);
}

void
mfbCopyRotatePixmap(register PixmapPtr psrcPix, 
		    register PixmapPtr *ppdstPix, 
		    int xrot, int yrot)
{
    register PixmapPtr pdstPix;
    int size, i;
    FBP psrc, pdst;

    if ((pdstPix = *ppdstPix) &&
	(pdstPix->devKind == psrcPix->devKind) &&
	(pdstPix->drawable.height == psrcPix->drawable.height))
    {
        psrc = (FBList *) psrcPix->devPrivate.ptr;
	pdst = (FBList *) pdstPix->devPrivate.ptr;
	size = psrcPix->drawable.height * (psrcPix->devKind >> 2);
	for (i = 0; i < size; ++ i) 
	  *(pdst ++) = *(psrc ++);
	pdstPix->drawable.width = psrcPix->drawable.width;
	pdstPix->drawable.serialNumber = NEXT_SERIAL_NUMBER;
    }
    else
    {
	if (pdstPix)
	    /* FIX XBUG 6168 */
	    (*pdstPix->drawable.pScreen->DestroyPixmap)(pdstPix);
	*ppdstPix = pdstPix = mfbCopyPixmap(psrcPix);
	if (!pdstPix)
	    return;
    }
    mfbPadPixmap(pdstPix);
    if (xrot)
	mfbXRotatePixmap(pdstPix, xrot);
    if (yrot)
	mfbYRotatePixmap(pdstPix, yrot);
}
